/*
 * NOTE: This example is works on x86 and powerpc.
 * Here's a sample kernel module showing the use of kprobes to dump a
 * stack trace and selected registers when do_fork() is called.
 *
 * For more information on theory of operation of kprobes, see
 * Documentation/kprobes.txt
 *
 * You will see the trace data in /var/log/messages and on the console
 * whenever do_fork() is invoked to create a new process.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kprobes.h>
#include <linux/limits.h> 

#include "probe_do_fork.h"

static char process_name[NAME_MAX];
//module_param_string(process, process_name, NAME_MAX, S_IRUGO);
//MODULE_PARM_DESC(process, "Process to monitor with kprobe");

/* For each probe you need to allocate a kprobe structure */
static struct kprobe kp = {
	.symbol_name	= "do_fork",
};

/* defined in mod_calc_stacks.h, implemented in mod_calc_stacks.c */
void signalThreadStart(pid_t, char*, long long);

void signalProgramNameProbeDoFork(char* p_name) {
	strlcpy(process_name, p_name, strlen(p_name));	
	printk (KERN_INFO "process_name in probe_do_fork.c set to %s\n", process_name);
}

/* kprobe pre_handler: called just before the probed instruction is executed */
static int handler_pre(struct kprobe *p, struct pt_regs *regs) {
	/* cur_p_name contains name current process */	
	//char cur_p_name[NAME_MAX];
	//strlcpy(cur_p_name, current->comm, sizeof(cur_p_name));	

	if ( strlen(process_name) > 0 && strcmp(current->comm,process_name) == 0 ) {
		signalThreadStart(current->pid, current->comm, (long long) ktime_to_ns(ktime_get()));
		//printk(KERN_INFO "new thread started at time: %lld, parent_pid: %d, p_name: %s\n", (long long) ktime_to_ns(ktime_get()), current->pid, cur_p_name);
	}

	return 0;
}

/* kprobe post_handler: called after the probed instruction is executed */
static void handler_post(struct kprobe *p, struct pt_regs *regs, unsigned long flags) { }

/*
 * fault_handler: this is called if an exception is generated for any
 * instruction within the pre- or post-handler, or when Kprobes
 * single-steps the probed instruction.
 */
static int handler_fault(struct kprobe *p, struct pt_regs *regs, int trapnr) {
	printk(KERN_INFO "fault_handler: p->addr = 0x%p, trap #%dn",
		p->addr, trapnr);
	/* Return 0 because we don't handle the fault. */
	return 0;
}

static int __init kprobe_init(void) {
	int ret;
	kp.pre_handler = handler_pre;
	kp.post_handler = handler_post;
	kp.fault_handler = handler_fault;

	ret = register_kprobe(&kp);
	if (ret < 0) {
		printk(KERN_INFO "register_kprobe failed, returned %d\n", ret);
		return ret;
	}
	printk(KERN_INFO "Planted kprobe at %p\n", kp.addr);
	return 0;
}

static void __exit kprobe_exit(void) {
	unregister_kprobe(&kp);
	printk(KERN_INFO "kprobe at %p unregistered\n", kp.addr);
}

/* Expose functions to other modules */
EXPORT_SYMBOL(signalProgramNameProbeDoFork); 

module_init(kprobe_init)
module_exit(kprobe_exit)
MODULE_LICENSE("GPL");
