/*
 * Probe that tracks when threads become active or scheduled out
 *
 * usage: insmod probe_sched_switch.ko process=<process_name>
 * parameter <process_name> is the name of the process we want to monitor,
 * default process is bash
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kprobes.h>
#include <linux/ktime.h>
#include <linux/limits.h>

#include "probe_sched_switch.h"

static char process_name[NAME_MAX];
//module_param_string(process, process_name, NAME_MAX, S_IRUGO);
//MODULE_PARM_DESC(process, "Process to monitor with probe");


void signalThreadScheduledOut(pid_t, char*, long long);
void signalThreadScheduledIn(pid_t, char*, long long);


/* function called when user writes process name in /proc/profiled_process */
void signalProgramNameProbeSchedSwitch(char* p_name) {
	strlcpy(process_name, p_name, strlen(p_name));	
	printk (KERN_INFO "process_name in probe_sched_switch.c set to %s\n", process_name);
}


/*
 * Jumper probe for probe_sched_switch function.
 * Mirror principle enables access to arguments of the probed function
 * from the probe handler.
 */

/* Proxy function having the same arguments as the actual probe_sched_switch function */

static void
jprobe_sched_switch(void *ignore, struct task_struct *prev, struct task_struct *next) {
	/* prev_p_name contains name of previous process */	
	//char prev_p_name[NAME_MAX];

	/* next_p_name contains name of previous process */	
	//char next_p_name[NAME_MAX];

	//strlcpy(prev_p_name, prev->comm, strlen(prev_p_name));	
	//strlcpy(next_p_name, next->comm, strlen(next_p_name));	

	if (strlen(process_name) > 0) {
		if ( strcmp(prev->comm,process_name) == 0 ) {
			signalThreadScheduledOut(prev->pid, prev->comm, (long long) ktime_to_ns(ktime_get()));		
			//printk(KERN_INFO "thread scheduled out at time: %lld, pid: %d, p_name: %s\n", (long long) ktime_to_ns(ktime_get()), prev->pid, prev_p_name);
		}

		if ( strcmp(next->comm,process_name) == 0 ) {
			signalThreadScheduledIn(next->pid, next->comm, (long long) ktime_to_ns(ktime_get()));	
			//printk(KERN_INFO "thread scheduled in at time: %lld, pid: %d, p_name: %s\n", (long long) ktime_to_ns(ktime_get()), next->pid, next_p_name);
		}
	}

	/* Always end with a call to jprobe_return(). */
	jprobe_return();
}

static struct jprobe my_jprobe = {
	.entry			= jprobe_sched_switch,
	.kp = {
		.symbol_name	= "probe_sched_switch",
	},
};

static int __init jprobe_init(void) {
	int ret;

	ret = register_jprobe(&my_jprobe);
	if (ret < 0) {
		printk(KERN_INFO "register_jprobe failed, returned %d\n", ret);
		return -1;
	}
	printk(KERN_INFO "Planted jprobe at %p, handler addr %p\n", my_jprobe.kp.addr, my_jprobe.entry);

	return 0;
}

static void __exit jprobe_exit(void) {
	unregister_jprobe(&my_jprobe);
	printk(KERN_INFO "jprobe at %p unregistered\n", my_jprobe.kp.addr);
}

/* Expose functions to other modules */
EXPORT_SYMBOL(signalProgramNameProbeSchedSwitch); 

module_init(jprobe_init)
module_exit(jprobe_exit)
MODULE_LICENSE("GPL");
